#ifndef IncludedGraphBase
#define IncludedGraphBase

#if GraphUnchecked
#define assert(x) 1
#endif

///////////////////////////////////////////////////////////////////////////////
// basic definitions

#include <stdio.h>
#include "objlist.h"
#include "intset.h"
#include "trie.h"

typedef unsigned int nat;
typedef enum {false, true} bool;

#define sqr(x) ((x)*(x))

#define NodeBits 13

///////////////////////////////////////////////////////////////////////////////
// graph representation: data structures

typedef enum labelType {
  LT_None,
  LT_Short, LT_Word, LT_Double, LT_Quad,
  LT_CString,
  LT_Illegal
} LabelType;

struct labelInfo {
  LabelType typ;
  nat nLabels;
  object *base, *aux;
};

typedef struct graph *Graph;

struct repShared {
  Graph graph;
  bool valid, sole;
  nat size;
};

typedef struct bitMatrixRep {
  struct repShared shared;
  BitMatrix *matrix;
} *BitMatrixRep;

typedef unsigned short LabelIndex;

typedef struct labelMatrixRep {
  struct repShared shared;
  LabelIndex *matrix;
} *LabelMatrixRep;

typedef unsigned int EdgeInfo;

typedef struct edgeArrayRep {
  struct repShared shared;
  int *offsets;
  EdgeInfo *base;
} *EdgeArrayRep;

struct graph {
  nat graphId;
  bool directed;
  nat modified;
  nat nNodes, nEdges, nSelfEdges, nNodesSeen, nEdgesSeen;
  struct labelInfo nodeLabels, edgeLabels;
  BitMatrixRep bitMatrixRep;
  LabelMatrixRep fullMatrixRep, symMatrixRep;
  EdgeArrayRep edgeArrayRep, outEdgeArrayRep;
  ObjList outEdgeListRep;
  ObjList crossIndices;
  int fileSz, *basePtr;
  FILE *file;
};

typedef int NodeId;

#define IllegalNode (-1)

///////////////////////////////////////////////////////////////////////////////
// error codes

typedef enum graphError {
  GE_None, GE_IOError, GE_MmapFailed, GE_Format, GE_OutOfMemory,
  GE_TooManyNodes, GE_TooManyEdges, GE_IncompleteGraph
} GraphError;

static GraphError lastGraphError = GE_None;
static char *graphErrorMessages[]; // use the error code as index in this array

#include "graphMacros.h"

///////////////////////////////////////////////////////////////////////////////
// graph creation

Graph newGraph (bool directed, LabelType nodeLabels, LabelType edgeLabels);

NodeId addNode (Graph g, object label);

#define AddNode(g,l) (addNode ((g), (object) (l)))

GraphError addEdge (Graph g, NodeId u, NodeId v, object label);

#define AddEdge(g,u,v,l) (addEdge ((g), (u), (v), (object) (l)))

Graph startKnownGraph (nat nNodes, nat nEdges, bool directed,
		       LabelType nodeLabels, LabelType edgeLabels);

GraphError endKnownGraph (Graph g);

///////////////////////////////////////////////////////////////////////////////
// representation-independent operations

LabelIndex getEdge (Graph g, NodeId i, NodeId j);

// graph iteration macros
//   (defined in graphMacro.h, this is for documentation purposes only)
#ifndef IncludedGraphMacros

/* ScanEdgesOnce :
 *   (Graph, var int, var int, var int, command, command, command) -> command
 * -- scan once each edge of the graph g
 *    in increasing lexicographical order of the (source,target) node pair;
 * -- pre is code run once for each source node, before scanning its neighbors
 * -- post is code run after scanning its neighbors
 * -- on is code run for each edge
 * -- i is the source node in pre, post, on
 * -- j and e are target node and edge label index, respectively, in on
 */
#define ScanEdgesOnce(g,i,j,e,pre,on,post)

/* ScanEdgesDirected :
 *   (Graph, var int, var int, var int, command, command, command) -> command
 * -- same as ScanEdgesOnce, except in undirected graphs non-self edges are
 *    visited in both directions
 */
#define ScanEdgesDirected(g,i,j,e,pre,on,post)

#endif

///////////////////////////////////////////////////////////////////////////////
// graph IO functions

GraphError openGraph (FILE *f, Graph g);
// expects the file to be open, and does not close it

GraphError closeGraph (Graph g);
// does not close the file associated with the graph

GraphError writeGraph (Graph g, FILE *f);
// expects the file to be open, and does not close it

///////////////////////////////////////////////////////////////////////////////
// basic parameters of a graph

// number of nodes
inline static nat nNodes (Graph g) {
  assert (g != NULL);
  return g->nNodes;
}

// number of edges
inline static nat nEdges (Graph g) {
  assert (g != NULL);
  return g->nEdges;
}

// number of self-edges
inline static nat nSelfEdges (Graph g) {
  assert (g != NULL);
  return g->nSelfEdges;
}

///////////////////////////////////////////////////////////////////////////////
// representation selection

BitMatrixRep bitMatrix (Graph g);

LabelMatrixRep fullMatrix (Graph g);

LabelMatrixRep symMatrix (Graph g);
// for undirected graphs only

EdgeArrayRep edgeArray (Graph g);

EdgeArrayRep outEdgeArray (Graph g);
// the same as edgeArray for directed graphs

///////////////////////////////////////////////////////////////////////////////
// label access functions

// auxiliary macro
#define GetLabel(t,lt,w,g,i) \
  (assert ((g)->w##Labels.typ == lt), \
   ((t) (g)->w##Labels.base) [i])

/* In all of these functions,
 * -- w is either the word "node" or the word "edge"
 * -- g is the graph representation (of type Graph)
 * -- i is the label index.
 * The result is the label for the corresponding entity (node/edge) at index i.
 */
#define LabelShort(w,g,i) GetLabel (short *, w, (g), (i))

#define LabelInt(w,g,i) GetLabel (int *, w, (g), (i))

#define LabelFloat(w,g,i) GetLabel (float *, w, (g), (i))

#define LabelDouble(w,g,i) GetLabel (double *, w, (g), (i))

#define LabelQuad(w,g,i) GetLabel (long double *, w, (g), (i))

#define LabelCString(w,g,i) GetLabel (char **, w, (g), (i))

///////////////////////////////////////////////////////////////////////////////
// auxiliary functions

// representation sizes
inline static nat symMtxSize (nat nRows) {
  return ((nRows + 1) * nRows) >> 1;
}

inline static nat fullMatrixRepSize (Graph g) {
  assert (g != NULL);
  return sqr (g->nNodes) * sizeof (LabelIndex) + sizeof (struct labelMatrixRep);
}

inline static nat symMatrixRepSize (Graph g) {
  assert (g != NULL && !g->directed);
  return symMtxSize (g->nNodes) * sizeof (LabelIndex)
    + sizeof (struct labelMatrixRep);
}

inline static nat outEdgeArrayRepSize (Graph g) {
  assert (g != NULL);
  return g->nNodes * sizeof (int) + g->nEdges * sizeof (EdgeInfo)
    + sizeof (struct edgeArrayRep);
}

inline static nat edgeArrayRepSize (Graph g) {
  assert (g != NULL);
  return g->directed
    ? outEdgeArrayRepSize (g)
    : g->nNodes * sizeof (int)
      + (g->nEdges * 2 - g->nSelfEdges) * sizeof (EdgeInfo)
      + sizeof (struct edgeArrayRep);
}

///////////////////////////////////////////////////////////////////////////////
// iterators on the matrix (using definitions from graphMacros.h)

// same meaning of the parameters as for ScanEdgesOnce

// scan the complete adjacency matrix
#define ScanMatrix(g,i,j,e,pre,on,post) \
  ScanGraphMtx (ScanMtxFull, fullMatrixRep, g, i, j, e, pre, on, post)

// scan the upper triangle of the adjacency matrix
#define ScanMatrixUpper(g,i,j,e,pre,on,post) \
  ScanGraphMtx (ScanMtxUpper, fullMatrixRep, g, i, j, e, pre, on, post)

// scan the complete adjacency matrix (for undirected graphs)
#define ScanFullMatrix(g,i,j,e,pre,on,post) \
  ScanMatrix (fullMatrix (g), i, j, e, pre, on, post)

///////////////////////////////////////////////////////////////////////////////
// iterators on the symmetric matrix representations

/* ScanSymMatirxUpper :
 *   (Graph, var int, var int, var int, command, command, command) -> command
 * -- scan once each edge of the undirected graph g
 *    in increasing lexicographical order of the (source,target) node pair;
 * -- pre is code run once for each source node, before scanning its neighbors
 * -- post is code run after scanning its neighbors
 * -- on is code run for each edge
 * -- i is the source node in pre, post, on
 * -- j and e are target node and edge label index, respectively, in on
 */
#define ScanSymMatrixUpper(g,i,j,e,pre,on,post) \
  ScanGraphMtx (ScanSymMtxUpper, symMatrixRep, g, i, j, e, pre, on, post)

/* ScanSymMatrix :
 *   (Graph, var int, var int, var int, command, command, command) -> command
 * -- like ScanSymMatrixUpper, except in undirected graphs non-self edges are
 *    visited in both directions
 */
#define ScanSymMatrix(g,i,j,e,pre,on,post) \
  ScanGraphMtx (ScanSymMtx, symMatrixRep, g, i, j, e, pre, on, post)

///////////////////////////////////////////////////////////////////////////////
// edge array access functions

/* descArrayStart :
 * -- get a pointer to the beginning of the array of descendants of node i
 */
inline static EdgeInfo *descArrayStart (EdgeArrayRep r, NodeId i) {
  assert (r != NULL && r->shared.valid);
  return r->base + r->offsets[i];
}

/* descArrayEnd :
 * -- get a pointer to the element after the end of the array of descendants
 *    of node i
 */
inline static EdgeInfo *descArrayEnd (EdgeArrayRep r, NodeId i) {
  return descArrayStart (r, i+1);
}

// edge info de-/construction for the edge array representation

// construct the edge info given the target node and the edge label index
inline static EdgeInfo edgeInfo (NodeId n, int v) {
  return (v << NodeBits) | n;
}

// extract the target node for the given edge
inline static NodeId edgeTarget (EdgeInfo e) {
  return e & ((1<<NodeBits)-1);
}

// extract the edge label index for the given edge
inline static int edgeValue (EdgeInfo e) {
  return e >> NodeBits;
}

inline static LabelIndex edgeArrayRepElt (EdgeArrayRep p, nat i, nat j) {
  assert (p != NULL && p->shared.valid);
  EdgeInfo *l, *r, *m, e = (EdgeInfo) -1;
  for (l = descArrayStart (p, i), r = descArrayEnd (p, i) - 1;
       l != r;
       j >= edgeTarget (e = *(m = l+((r-l) / 2))) ? (r = m) : (l = m+1));
  return j == edgeTarget (e) ? edgeValue (e) + 1 : 0;
}

inline static LabelIndex outEdgeArrayRepElt (EdgeArrayRep r, nat i, nat j) {
  assert (r != NULL && r->shared.graph != NULL);
  if (!r->shared.graph->directed && i > j) { nat t = i; i = j; j = t; }
  return edgeArrayRepElt (r, i, j);
}

// iterators on the edge array representation

/* ScanDescendantsOf :
 *   (Graph, int, var int, var int, command) -> command
 * -- scan the immediate descendants of a node
 * -- i is the node
 * -- on is code run once for each edge
 * -- j is the current descendant node, available in on
 * -- e is the edge label index, available in on
 */
#define ScanDescendantsOf(g,i,j,e,on) \
  ScanEdgeArrayOf (outEdgeArray (g), i, j, e, on)

/* ScanNeighborsOf :
 *   (Graph, int, var int, var int, command) -> command
 * -- same as ScanDescendantsOf, but for all neighbors in undirected graphs
 */
#define ScanNeighborsOf(g,i,j,e,on) \
  ScanEdgeArrayOf (edgeArray (g), i, j, e, on)

/* ScanEdgeArrayOnce :
 *   (Graph, var int, var int, var int, command, command, command) -> command
 * -- same as ScanEdgeArray, but only the outgoing edges in a directed graph
 *    or each edge once in a undirected,
 *    and given a Graph instead of an EdgeArrayRep
 */
#define ScanEdgeArrayOnce(g,i,j,e,pre,on,post) \
  ScanEdgeArray (outEdgeArray (g), i, j, e, pre, on, post)

/* ScanEdgeArrayFull :
 *   (Graph, var int, var int, var int, command, command, command) -> command
 * -- same as ScanEdgeArrayOnce, but for undirected graphs non-self edges are
 *    visited in both directions
 */
#define ScanEdgeArrayFull(g,i,j,e,pre,on,post) \
  ScanEdgeArray (edgeArray (g), i, j, e, pre, on, post)

#endif
