#ifndef IncludedGraphMacros
#define IncludedGraphMacros

///////////////////////////////////////////////////////////////////////////////
/* The primitive access functions do not check if the graph representation
 * is the one they expect. Since the representation determines to some extent
 * which algorithms are better suited, this check should be done at the
 * algorithm selection level.
 */

///////////////////////////////////////////////////////////////////////////////
// symmetric adjacency matrix access functions

inline static nat symMtxRowBase_unchkd (nat nRows, nat row) {
  return (row * (nRows * 2 - 1 - row)) >> 1;
}

inline static nat symMtxRowBase (nat nRows, nat row) {
  assert (row < nRows);
  return symMtxRowBase_unchkd (nRows, row);
}

inline static nat symMtxUpperCell (nat nRows, nat row, nat column) {
  assert (row <= column && column < nRows);
  return symMtxRowBase_unchkd (nRows, row) + column;
}

inline static nat symMtxCell (nat nRows, nat row, nat column) {
  return row <= column
    ? symMtxUpperCell (nRows, row, column)
    : symMtxUpperCell (nRows, column, row);
}

/* -- computes the address of the label of the edge between nodes i and j
 * -- value in cell :  0 -> no edge, 1+n -> edge label index n
 */
static inline LabelIndex *symMatrixRepElt (LabelMatrixRep r, nat i, nat j) {
  assert (r != NULL && r->shared.valid);
  return &r->matrix[symMtxCell (r->shared.graph->nNodes, i, j)];
}

///////////////////////////////////////////////////////////////////////////////
// adjacency matrix access functions

/* -- computes the address of the label of the edge between nodes i and j
 * -- value in cell:  0 -> no edge, 1+n -> edge label index n
 */
static inline LabelIndex *matrixRepElt (LabelMatrixRep r, nat i, nat j) {
  nat n = r->shared.graph->nNodes;
  assert (r != NULL && r->shared.valid && i < n && j < n);
  return &r->matrix[i*n+j];
}

///////////////////////////////////////////////////////////////////////////////
// iterators

#define ScanEdgesOnce(g,i,j,e,pre,on,post) do { \
  Graph __graph = (g); \
  assert (__graph != NULL && !__graph->directed); \
  EdgeArrayRep __ear = __graph->outEdgeArrayRep; \
  if (__ear != NULL && __ear->shared.valid) \
    ScanEdgeArray (__ear, i, j, e, pre, on, post); \
  else { \
    LabelMatrixRep __amr = __graph->symMatrixRep; \
    if (__amr != NULL && __amr->shared.valid) \
      ScanSymMatrixUpper (__graph, i, j, e, pre, on, post); \
    else ScanEdgeArrayOnce (__graph, i, j, e, pre, on, post); \
  } \
} while (0)

#define ScanEdgesDirected(g,i,j,e,pre,on,post) do { \
  Graph __graph = (g); \
  assert (__graph != NULL); \
  EdgeArrayRep __ear = __graph->edgeArrayRep; \
  if (__ear != NULL && __ear->shared.valid) \
    ScanEdgeArray (__ear, i, j, e, pre, on, post); \
  else { \
    LabelMatrixRep __amr = __graph->fullMatrixRep; \
    if (__amr != NULL && __amr->shared.valid) \
      ScanMtxFull (__amr->matrix, __graph->nNodes, i, j, e, pre, on, post); \
    else { \
      __amr = __graph->symMatrixRep; \
      if (__amr != NULL && __amr->shared.valid) \
	 ScanSymMtxFull (__amr->matrix, __graph->nNodes, i, j, e, pre, on, post); \
      else if (edgeArrayRepSize (__graph) < fullMatrixRepSize (__graph)) \
	ScanEdgeArrayFull (__graph, i, j, e, pre, on, post); \
      else ScanMtxFull (fullMatrix (__graph), i, j, e, pre, on, post); \
    } \
  } \
} while (0)

#define ScanMtx(mtx,n,i,j,v,pre,on,post,firstColumn) do { \
  int __mtx_nodes = (n); \
  LabelIndex *__mtx = (mtx); \
  for (i = 0; i < __mtx_nodes; ++i) { \
    pre \
    for (j = firstColumn; j < __mtx_nodes; ++j, ++__mtx) { \
      v = *__mtx; \
      on \
    } \
    post \
  } \
} while (0)

#define ScanMtxFull(mtx,n,i,j,v,pre,on,post) \
  ScanMtx (mtx, n, i, j, v, pre, on, post, 0)

#define ScanMtxUpper(mtx,n,i,j,v,pre,on,post) do { \
  int __mtx_nodes = (n); \
  LabelIndex *__mtx = (mtx); \
  for (i = 0; i < __mtx_nodes; ++i, __mtx += i) { \
    pre \
    for (j = i; j < __mtx_nodes; ++j, ++__mtx) { \
      v = *__mtx; \
      on \
    } \
    post \
  } \
} while (0)

// auxiliary higher-order macro (graph-to-matrix adapter)
#define ScanGraphMtx(core,rep,g,i,j,e,pre,on,post) do { \
    LabelMatrix __mtx_rep = (g)->rep; \
    assert (__mtx_rep != NULL); \
    core (__mtx_rep->matrix, (g)->nNodes, i, j, e, pre, \
          { if (e) { --e; on } }, post); \
  } while (0)

/* ScanSymMtxUpper :
 *   (*EdgeInfo, int, var NodeId, var NodeId, var EdgeInfo,
 *    command, command, command)
 *   -> command
 * -- scan the upper triangle of the symmetrix matrix m, column first
 * -- n is the number of rows/columns in m
 * -- pre is code run once before scanning each row
 * -- post is code run once after scanning each row
 * -- on is code run for each cell
 * -- i is the row currently scanned, available in pre, post, and on
 * -- j is the column currently scanned, available in on
 * -- v is the value of the cell m(i,j), available in on
 */
#define ScanSymMtxUpper(mtx,n,i,j,v,pre,on,post) \
  ScanMtx (mtx, n, i, j, v, pre, on, post, i)

/* ScanSymMtx :
 *   (*EdgeInfo, int, var NodeId, var NodeId, var EdgeInfo,
 *    command, command, command)
 *   -> command
 * -- like ScanSymMtxUpper, except the complete matrix is scanned
 */
#define ScanSymMtx(mtx,n,i,j,v,pre,on,post) do { \
  int __mtx_nodes = (n), __mtx_rowbase = 0; \
  for (i = 0; i < __mtx_nodes; ++i, __mtx_rowbase += __mtx_nodes-i) { \
    pre \
    int __mtx_colbase = 0; \
    for (j = 0; j < i; ++j, __mtx_colbase += __mtx_nodes-j) { \
      v = (mtx)[i+__mtx_colbase]; \
      on \
    } \
    for (; j < __mtx_nodes; ++j) { \
      v = (mtx)[__mtx_rowbase+j]; \
      on \
    } \
    post \
  } \
} while (0)

/* ScanEdgeArrayOf :
 *   (EdgeArrayRep, NodeId, var NodeId, var int, command) -> command
 * -- scan the immediate descendants of a node
 * -- i is the node
 * -- on is code run once for each edge
 * -- j is the current descendant node, available in on
 * -- e is the edge label index, available in on
 */
#define ScanEdgeArrayOf(r,i,j,e,on) do { \
  EdgeArrayRep __seq_rep = (r); \
  assert (__seq_rep != NULL); \
  int *__edge_ptr = descArrayStart (__seq_rep, (i)), \
      *__seq_end = descArrayEnd (__seq_rep, (i)); \
  while (__edge_ptr < __seq_end) { \
    e = *__edge_ptr; \
    j = edgeTarget (e); \
    e = edgeValue (e); \
    on \
    ++__edge_ptr; \
  } \
} while (0)

/* ScanEdgeArray :
 *   (EdgeArrayRep, var NodeId, var NodeId, var int, command, command, command)
 *    -> command
 * -- scan once each edge of the graph represented by r
 *    in increasing lexicographical order of the (source,target) node pair;
 * -- pre is code run once for each source node, before scanning its neighbors
 * -- post is code run after scanning its neighbors
 * -- on is code run for each edge
 * -- i is the source node in pre, post, on
 * -- j and e are target node and edge label index, respectively, in on
 */
#define ScanEdgeArray(r,i,j,e,pre,on,post) do { \
  EdgeArrayRep __list_rep = (r); \
  int __graph_nodes = __list_rep->shared.graph->nNodes; \
  int *__seq_end = __list_rep->base; \
  for (i = 0; i < __graph_nodes; ++i) { \
    int *__edge_ptr = __seq_end; \
    __seq_end = descArrayEnd (i); \
    pre \
    while (__edge_ptr < __seq_end) { \
      e = *__edge_ptr; \
      j = edgeTarget (e); \
      e = edgeValue (e); \
      on \
      ++__edge_ptr; \
    } \
    post \
  } \
} while (0)

#endif
