#include "inthacks.h"
#include "intset.h"
#include "objlist.h"

#define MemChunkSz (1<<18)

#if DEBUG
#define debug(x...) printf (x), fflush (stdout)
#else
#define debug(x...) 1
#endif

inline static int nodeDegree (BitMatrix graph, int nNodes, int i) {
  return SetSize (BitMatrixRow (graph, nNodes, i), nNodes);
}

inline static int maxDegree (BitMatrix graph, int nNodes) {
  int i, d = 0;
  for (i = 0; i < nNodes; ++i)
    d = intMax (d, nodeDegree (graph, nNodes, i));
  return d;
}

typedef struct cliqTree *CliqTree;
struct cliqTree {
  CliqTree with, without, up;
  int node;
};

struct graphInfo *graphInfo;

struct maxCliques_status {
#if OBSERVED
  int node, memory;
#endif
#if STATS
  int curNode, extends, cells, topins, searches, steps, dups, topdups;
#endif
} maxCliques_status;

FILE *maxCliques_traceout;

int maxCliqueSize;
#if LARGEST_ONLY
ObjList largest, *endLargest;
#endif

#if STATS
#define bump(x) (++maxCliques_status.x)
#else
#define bump(x)
#endif

#define ScanCliqTree(root,cursor,top,enter,complete,exit) { \
  cursor = root; \
  do { \
    if (cursor->up == NULL) top; \
    do \
      enter \
    while (cursor->with != NULL && (cursor = cursor->with, 1)); \
    complete; \
    do \
      exit \
    while (cursor->without == NULL && (cursor = cursor->up) != NULL); \
  } while (cursor != NULL && (cursor = cursor->without, 1)); \
}

inline static void printClique (CliqTree p, int running) {
  for (; running ? p->up : p; p = p->up)
#if DEBUG
    fprintf (maxCliques_traceout, " %d", intAbs (p->node));
#else
    fprintf (maxCliques_traceout, " %s", graphInfo->names[intAbs (p->node)]);
#endif
  fprintf (maxCliques_traceout, "\n");
}

#if LARGEST_ONLY
inline static void printLargest () {
  ObjList *l;
  for (l = &largest; l != endLargest; l = &(*l)->tail)
    printClique ((CliqTree) (*l)->head, 0);
}
#endif

inline static void trackLargest (CliqTree q, int sz) {
#if LARGEST_ONLY
  if (sz >= maxCliqueSize) {
    debug (" %d>=%d", sz, maxCliqueSize);
    fprintf (maxCliques_traceout, "%d:", sz);
    printClique (q, 1);
    fflush (maxCliques_traceout);
#endif
    if (sz > maxCliqueSize) {
      debug (">");
      maxCliqueSize = sz;
#if LARGEST_ONLY
      endLargest = &largest;
#endif
    }
#if LARGEST_ONLY
    ObjList p = *endLargest;
    if (p == NULL) {
      *endLargest = p = New (1, struct objList);
      p->tail = NULL;
    }
    p->head = (object) q;
    endLargest = &p->tail;
  }
#endif
  bump (extends);
}

ObjList arenaList = NULL;
CliqTree arena;
int freeIndex = MemChunkSz;

void freeCliqSpace () {
#if ALLOC_BIG_CHUNKS
  FreeObjList (arenaList, free);
  arenaList = NULL;
  freeIndex = MemChunkSz;
#endif
}

inline static CliqTree
newCliqTree (CliqTree *at, int node, CliqTree without, CliqTree up) {
#if ALLOC_BIG_CHUNKS
  if (freeIndex == MemChunkSz) {
    //    fprintf (stderr, "#");
    debug ("#");
    arena = New (MemChunkSz, struct cliqTree);
#if OBSERVED
    maxCliques_status.memory += MemChunkSz * sizeof (struct cliqTree);
#endif
    ConsOn (arenaList, arena);
    freeIndex = 0;
  }
  CliqTree z = &arena[freeIndex++];
#else
  CliqTree z = New (1, struct cliqTree);
#if OBSERVED
  maxCliques_status.memory += sizeof (struct cliqTree) + 8; // ?
#endif
#endif
  *at = z;
  z->node = node;
  z->without = without;
  z->up = up;
  bump (cells);
  return z;
}

struct nodeStackEntry {
  int node;
  IntSet ns;
};

CliqTree maxCliques (struct graphInfo *g) {
  int i, j;
  int *candPtr, cliqSz, validNS;
  CliqTree c, p, *path, *previous;
  struct cliqTree root;
  IntSet iNeighbors, cns;
  struct nodeStackEntry *candStk;
  BitMatrix graph = g->graph;
  int nNodes = g->nNodes;
  graphInfo = g;
#if OBSERVED
  maxCliques_status.memory = maxCliques_status.node = 0;
#endif
#if STATS
  memset ((void *) &maxCliques_status, 0, sizeof (struct maxCliques_status));
#endif
  maxCliqueSize = 0;
#if LARGEST_ONLY
  largest = NULL;
  endLargest = &largest;
  fflush (maxCliques_traceout);
#endif
  {
    int maxCliqSz = maxDegree (graph, nNodes) +1;
    debug ("maxCliqSz = %d\n", maxCliqSz);
    candStk = New (maxCliqSz, struct nodeStackEntry);
    for (i = 1; i < maxCliqSz; ++i) candStk[i].ns = NewSet (nNodes);
    candPtr = New (maxCliqSz, int);
    path = New (maxCliqSz, CliqTree);
    root.with = root.without = root.up = NULL;
#if DEBUG
    root.node = -1;
#endif
  }
  for (i = 0; i < nNodes; ++i) {
#if TRACKING
    fprintf (maxCliques_traceout, "%d %d %d %d %d %d %d %d %d\n",
	     i, maxCliqueSize, maxCliques_status.cells,
	     maxCliques_status.extends,  maxCliques_status.topins,
	     maxCliques_status.dups, maxCliques_status.topdups,
	     maxCliques_status.searches, maxCliques_status.steps);
    fflush (maxCliques_traceout);
#else
    if (i % 100 == 0) fprintf (stderr, ".");
#endif
#if OBSERVED
    maxCliques_status.node = i;
#endif
    candStk[0].ns = iNeighbors = BitMatrixRow (graph, nNodes, i);
    for (p = &root, c = p->with, cliqSz = validNS = 0;;) {
      debug (" _%d", p->node);
      if (c == NULL) {
	debug (" +%d", i);
	CliqTree q = newCliqTree (&p->with, i, NULL, p);
	q->with = NULL;
	trackLargest (q, cliqSz+1);
      moveup:
	c = p;
	p = p->up;
	if (--cliqSz < 0) break;
      }
      else if ((j = c->node, debug (" @%d", j), j) < 0) {
	debug (" inv");
	c->node = -j;
      }
      else if (Member (iNeighbors, j)) {
	++cliqSz;
	debug (" V[%d]", cliqSz);
	p = c;
	c = p->with;
	continue;
      }
      else {
	debug (" *[%d;%d]", cliqSz, validNS);
	int candNodeNdx, nodeNdx;
	CliqTree t = p, d;
	cns = candStk[validNS].ns;
	if (cliqSz == 0 || validNS < cliqSz) {
	  int empty;
	  for (nodeNdx = cliqSz; nodeNdx > validNS; ) {
	    debug ("..%d", t->node);
	    path[--nodeNdx] = t;
	    t = t->up;
	  }
	  if (cliqSz == 0) {
	    debug (" O");
	    empty = IsEmpty (cns, i);
	  }
	  else {
	    IntSet cns1;
	    while (validNS < cliqSz-1) {
	      debug (" X");
	      cns1 = candStk[validNS+1].ns;
	      // assuming no self-loops
	      SetIntersect (cns1, cns,
			    BitMatrixRow (graph, nNodes, path[validNS++]->node),
			    i);
	      cns = cns1;
	    }
	    debug (" Z");
	    cns1 = candStk[validNS+1].ns;
	    empty = SetIntersectEmpty
	      (cns1, cns, BitMatrixRow (graph, nNodes, path[validNS++]->node), i);
	    cns = cns1;
	  }
	  debug (" got-set");
	  if (empty) {
	    debug (" top-ins(%d)", i);
	    bump (topins);
	    for (t = c;
		 (d = t->without) != NULL && (j = intAbs (d->node)) < i;
		 t = d);
	    if (d == NULL || j > i) {
	      debug ("after(%d)", t->node);
	      CliqTree q = newCliqTree (&t->without, i, d, p);
	      q->with = NULL;
	      trackLargest (q, cliqSz+1);
	    }
	    else {
	      debug (" already");
	      bump (topdups);
	      d->node = i;
	    }
	    goto moveup;
	  }
	}
	debug (" search[%d]", cliqSz);
	for (t = c, d = c->with, candNodeNdx = cliqSz, nodeNdx = 0;;) {
	  debug (" __%d", t->node);
	  if (d == NULL) {
	    debug (" up");
	    d = t;
	    t = t->up;
	    if (--nodeNdx < 0) break;
	  }
	  else {
	    debug (" @@%d[%d,%d]", d->node, nodeNdx, candNodeNdx);
	    candPtr[nodeNdx++] = candNodeNdx;
	    j = d->node;
	    if (j < 0) {
	      d->node = -j;
	    }
	    else if (Member (iNeighbors, j)) {
	      debug (" in");
	      IntSet cns1 = candStk[++candNodeNdx].ns;
	      candStk[candNodeNdx].node = j;
	      debug ("-");
	      if (SetIntersectEmpty (cns1, cns,
				     BitMatrixRow (graph, nNodes, j),
				     i)) {
		CliqTree q = p, r, *s;
		candStk[++candNodeNdx].node = i;
		int si = cliqSz, k = candStk[++si].node, newBr = 0;
		debug (" empty[%d,%d]", c->node, k);
		bump (searches);
		for (s = &c->without;;) {
		  r = *s;
		  if (r == NULL || (j = intAbs (r->node), j > k)) {
		    debug (" insert%d(%d", q->node, candStk[si].node);
		    r = newCliqTree (s, signFlip (-k, newBr), r, q);
		    while (++si <= candNodeNdx) {
		      debug ("+%d", candStk[si].node);
		      r = newCliqTree (&r->with, candStk[si].node, NULL, r);
		    }
		    r->with = NULL;
		    trackLargest (r, candNodeNdx);
		    debug (")");
		    break;
		  }
		  else if (j < k) {
		    debug (">%d", r->node);
		    bump (steps);
		    s = &r->without;
		  }
		  else {
		    debug ("|%d", r->node);
		    if (++si > candNodeNdx) {
		      debug (" match");
		      bump (dups);
		      break;
		    }
		    k = candStk[si].node;
		    newBr |= signFill (r->node);
		    q = r;
		    s = &q->with;
		    bump (searches);
		  }
		}
		candNodeNdx -= 2;
		--nodeNdx;
	      }
	      else {
		debug (" nonempty");
		cns = cns1;
		t = d;
		d = t->with;
		continue;
	      }
	    }
	    else {
	      debug (" notin");
	      t = d;
	      d = t->with;
	      continue;
	    }
	  }
	  if ((d = d->without) == NULL) {
	    debug (" pull[%d]", nodeNdx);
	    while (nodeNdx-- >= 0 && (d = t->without, t = t->up, d == NULL))
	      debug ("^");
	    if (nodeNdx < 0) break;
	  }
	  candNodeNdx = candPtr[nodeNdx];
	  cns = candStk[candNodeNdx].ns;
	  debug (" end-search");
	}
	debug (" end-new");
      }
      if ((c = c->without) == NULL) {
	debug (" pop");
	while (cliqSz-- >= 0 && (c = p->without, p = p->up, c == NULL))
	  debug ("^");
	if (cliqSz < 0) break;
      }
      validNS = intMin (validNS, cliqSz);
      debug (" vns=%d", validNS);
    }
    debug (" phase\n");
#if DEBUG
    printCliqTree (root.with);
#endif
  }
  debug ("done\n");
#if !TRACKING
  fprintf (stderr, "\n");
#endif
#if STATS
  printf (//stderr,
	   "\nmax size %d\n%d cells, %d extensions, %d top inserts\n%d inner duplicates, %d top duplicates\n%d searches, %d search steps\n",
	   maxCliqueSize,
	   maxCliques_status.cells, maxCliques_status.extends,
	   maxCliques_status.topins,
	   maxCliques_status.dups, maxCliques_status.topdups,
	   maxCliques_status.searches, maxCliques_status.steps);
#endif
  for (p = root.with; p; p = p->without) p->up = NULL;
  return root.with;
}
